#include <stdio.h>
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#define PIXVAL(iplimagep, x, y) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (x)))
#define PIXVALC(iplimagep, x, y, c) (*(uchar *)((iplimagep)->imageData + (y) * (iplimagep)->widthStep + (x) * 3 + (c)))

//------------------------------------------------------------------ 

enum track_state { TRACK_OFF, TRACK_INIT, TRACK_ON, TRACK_SEL };

struct mouse_param {
    CvBox2D *tboxp;
    enum track_state *tstatep;
};

void on_mouse(int event, int x, int y, int flags, void *param)
{
    static CvPoint topleft;

    struct mouse_param *mpp = (struct mouse_param *)param;

    if (*(mpp->tstatep) == TRACK_SEL) {
	if (x > topleft.x && y > topleft.y) {
	    mpp->tboxp->center = cvPoint2D32f((x + topleft.x) / 2,
					      (y + topleft.y) / 2);
	    mpp->tboxp->size = cvSize2D32f(x - topleft.x, y - topleft.y);
	} else {
	    mpp->tboxp->center = cvPoint2D32f(topleft.x, topleft.y);
	    mpp->tboxp->size = cvSize2D32f(0, 0);
	}
    }

    switch (event) {
    case CV_EVENT_LBUTTONDOWN:
	topleft = cvPoint(x, y);
	mpp->tboxp->center = cvPoint2D32f(x, y);
	mpp->tboxp->size = cvSize2D32f(0, 0);
	mpp->tboxp->angle = 90.0;
	*(mpp->tstatep) = TRACK_SEL;
	break;
    case CV_EVENT_LBUTTONUP:
	if (mpp->tboxp->size.width > 0 && mpp->tboxp->size.height > 0) {
	    *(mpp->tstatep) = TRACK_INIT;
	} else {
	    *(mpp->tstatep) = TRACK_OFF;
	}
	break;
    }
}

//------------------------------------------------------------------ 
#define R_BINS 16
#define G_BINS 16
#define B_BINS 16
#define BINS (R_BINS * G_BINS * B_BINS)

#define HISTVAL(hist, r, g, b) hist[(r) + ((g) * R_BINS) + ((b) * R_BINS * G_BINS)]

float *create_hist()
{
    return (float *)malloc(sizeof(float) * BINS);
}

void release_hist(float *hist)
{
    free(hist);
}

void calc_hist(IplImage *img, CvBox2D *tboxp, float *hist,
	       enum track_state *tsp)
{
    int i, j;
    int ww = (int)(tboxp->size.width / 2);
    int hh = (int)(tboxp->size.height / 2);
    int cx = (int)(tboxp->center.x);
    int cy = (int)(tboxp->center.y);
    int imin = MAX(cx - ww, 0);
    int imax = MIN(cx + ww, img->width - 1);
    int jmin = MAX(cy - hh, 0);
    int jmax = MIN(cy + hh, img->height - 1);
    float sum = 0;
    float w_normalizer = 2 / tboxp->size.width;
    float h_normalizer = 2 / tboxp->size.height;

    memset(hist, 0, sizeof(float) * BINS);

    for (j = jmin; j <= jmax; j++) {
	for (i = imin; i <= imax; i++) {
	    float x_norm = (i - cx) * w_normalizer;
	    float y_norm = (j - cy) * h_normalizer;
	    float r2 = x_norm * x_norm + y_norm * y_norm;
	    float weight;
	    if (r2 <= 1) { // Epanechnikov profile (up to scales)
		weight = 1 - r2;
	    } else {
		weight = 0;
	    }

	    uchar bidx = PIXVALC(img, i, j, 0) / (256 / B_BINS);
	    uchar gidx = PIXVALC(img, i, j, 1) / (256 / G_BINS);
	    uchar ridx = PIXVALC(img, i, j, 2) / (256 / R_BINS);
	    
	    HISTVAL(hist, ridx, gidx, bidx) += weight;
	    sum += weight;
	}
    }
    if (sum == 0) {
	*tsp = TRACK_OFF;
	return;
    }
    for (i = 0; i < BINS; i++) {
	hist[i] /= sum;
    }
}

float bhattacharyya_coeff(float *hist1, float *hist2)
{
    int i;
    float bhatt_coeff = 0;

    for (i = 0; i < BINS; i++) {
	bhatt_coeff += sqrt(hist1[i] * hist2[i]);
    }

    return bhatt_coeff;
}

//------------------------------------------------------------------ 

#define MAX_ITER 20

float mean_shift(IplImage *img, CvBox2D *tboxp, 
		 float *hist_model, float *hist_cand, enum track_state *tsp)
{
    float prev_cx, prev_cy;
    float displace2;
    int ww = (int)(tboxp->size.width / 2);
    int hh = (int)(tboxp->size.height / 2);
    int iter = 0;

    do {
	int i, j;
	int cx = (int)(tboxp->center.x);
	int cy = (int)(tboxp->center.y);
	int imin = MAX(cx - ww, 0);
	int imax = MIN(cx + ww, img->width - 1);
	int jmin = MAX(cy - hh, 0);
	int jmax = MIN(cy + hh, img->height - 1);
	float m0 = 0, mx = 0, my = 0;
	prev_cx = tboxp->center.x;
	prev_cy = tboxp->center.y;

	calc_hist(img, tboxp, hist_cand, tsp);

	if (*tsp == TRACK_OFF) {
	    return 0;
	}

	for (j = jmin; j <= jmax; j++) {
	    for (i = imin; i <= imax; i++) {
		uchar bidx = PIXVALC(img, i, j, 0) / (256 / B_BINS);
		uchar gidx = PIXVALC(img, i, j, 1) / (256 / G_BINS);
		uchar ridx = PIXVALC(img, i, j, 2) / (256 / R_BINS);
		float q = HISTVAL(hist_model, ridx, gidx, bidx);
		float p = HISTVAL(hist_cand, ridx, gidx, bidx);
		if (p > 0) {
		    float w = sqrt(q / p);
		    m0 += w;
		    mx += w * i;
		    my += w * j;
		}
	    }
	}
	if (m0 == 0) {
	    *tsp = TRACK_OFF;
	    return 0;
	}
	tboxp->center.x = mx / m0;
	tboxp->center.y = my / m0;

	displace2 = (tboxp->center.x - prev_cx) * (tboxp->center.x - prev_cx)
	    + (tboxp->center.y - prev_cy) * (tboxp->center.y - prev_cy);
    } while (++iter < MAX_ITER && displace2 >= 3);

    return bhattacharyya_coeff(hist_model, hist_cand);
}

//------------------------------------------------------------------ 

int main()
{
    int is_size_adaptive = 0;
    int key;
    float *hist_model, *hist_cand;
	
    hist_model = create_hist();
    hist_cand = create_hist();

    IplImage *frame, *img;
    CvCapture *capture = cvCaptureFromCAM(0);
    if (!capture) {
	fprintf(stderr, "cannot capture\n");
	return 1;
    }

    frame = cvQueryFrame(capture);
    img = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);

    enum track_state tstate;
    CvBox2D tbox;
    struct mouse_param mparam = { &tbox, &tstate };

    cvNamedWindow("track", CV_WINDOW_AUTOSIZE);
    cvSetMouseCallback("track", on_mouse, &mparam);

    printf("q: quit, s: toggle size adaptive/fixed\n");
    
    while (1) {
	frame = cvQueryFrame(capture); /* cannot be modified */
	cvFlip(frame, img, NULL);

	if (tstate == TRACK_INIT) {
	    calc_hist(img, &tbox, hist_model, &tstate);
	    tstate = TRACK_ON;
	} else if (tstate == TRACK_SEL
		   && tbox.size.width > 0 && tbox.size.height > 0) {
	    cvEllipseBox(img, tbox, CV_RGB(255, 0, 0), 3);
	} else if (tstate == TRACK_ON) {
	    float bl, bs;
	    float b;
	    CvBox2D tl, ts;
            if (is_size_adaptive) {
                tl = ts = tbox;
                tl.size.width *= 1.1f;
                tl.size.height *= 1.1f;
                ts.size.width *= 0.9f;
                ts.size.height *= 0.9f;
            }

	    b = mean_shift(img, &tbox, hist_model, hist_cand, &tstate);

            if (is_size_adaptive) {
                bl = mean_shift(img, &tl, hist_model, hist_cand, &tstate);
                bs = mean_shift(img, &ts, hist_model, hist_cand, &tstate);
                if (bl > b && bl >= bs) {
                    b = bl;
                    tbox.size.width *= 1.03f;
                    tbox.size.height *= 1.03f;
                } else if (bs > b && bs > bl) {
                    b = bs;
                    tbox.size.width *= 0.97f;
                    tbox.size.height *= 0.97f;
                }
            }

	    if (b > 0.5) {
		cvEllipseBox(img, tbox, CV_RGB(255, 0, 0), 3);
	    } else {
		cvEllipseBox(img, tbox, CV_RGB(0, 255, 0), 1);
	    }
	}
	cvShowImage("track", img);
	if ((key = cvWaitKey(30)) > 0) {
            if (key == 'q' || key == 0x1b) {
                break;
            } else if (key == 's')  {
                is_size_adaptive = 1 - is_size_adaptive;
                printf(is_size_adaptive? "size adaptive\n": "size fixed\n");
            }
	}
    }

    return 0;
}
